﻿using HSolutions.Framework.MicroServices.WindowsService;
using Microsoft.Owin.Hosting;
using System;
using System.Reflection;
using System.ServiceProcess;
using System.Linq;

namespace HSolutions.Framework.MicroServices
{
    public abstract class MicroServiceBase : IDisposable
    {
        #region Events

        public event Action OnServiceStarted;
        public event Action OnServiceStopped;

        #endregion

        #region Fields

        internal readonly string _serviceDisplayName;
        internal readonly string _host;
        internal readonly int    _port;
        internal readonly WindowsServiceManager _serviceManager;
        internal readonly RegistryManipulator _registryManipulator;

        internal IDisposable _dispose;

        #endregion

        #region C'tor

        public MicroServiceBase(int port = 9000, string host = "http://localhost", string serviceDisplayName = null, string serviceName = null)
        {
            _host = host;
            _port = port;

            var assemblyName = Assembly.GetEntryAssembly().GetName().Name;
            _serviceDisplayName = serviceDisplayName ?? assemblyName;
            serviceName = serviceName ?? assemblyName;

            _serviceManager = new WindowsServiceManager(_serviceDisplayName);
            _registryManipulator = new RegistryManipulator(serviceName);

            InternalService.OsStarted += Start;
            InternalService.OsStopped += Stop;
            ProjectInstaller.InitInstaller(_serviceDisplayName, serviceName);

        }

        #endregion

        #region Public

        public void Run(string[] args)
        {
            if (!args.Any())
            {
                RunConsole();
                return;
            }

            switch (args[0])
            {
                case "-service":
                    RunService();
                    break;
                case "-install":
                    InstallService();
                    break;
                case "-uninstall":
                    UnInstallService();
                    break;
                default:
                    throw new Exception(args[0] + " comando inválido!!");
            }
        }

        #endregion

        #region Private

        private void RunConsole()
        {
            Start();

            Console.WriteLine("[{2}] Serviço {0} iniciado na porta {1}", _serviceDisplayName, _port, DateTime.Now);
            Console.WriteLine("Pressione qualquer tecla para encerrar este processo...");
            Console.ReadKey();

            Stop();
        }

        private static void RunService()
        {
            ServiceBase[] servicesToRun = { new InternalService() };
            ServiceBase.Run(servicesToRun);
        }

        private void InstallService()
        {
            _serviceManager.Install();
            _registryManipulator.AddMinusServiceToRegistry();
            _serviceManager.Start();
        }

        private void UnInstallService()
        {
            _serviceManager.Stop();
            _registryManipulator.RemoveMinusServiceFromRegistry();
            _serviceManager.UnInstall();
        }

        private void Stop()
        {
            this.Dispose();

            if (OnServiceStopped != null)
                OnServiceStopped.Invoke();
        }

        internal virtual void Start()
        {
            _dispose = WebApp.Start<StartupBase>(string.Concat(_host, ":", _port));
            if (OnServiceStarted != null)
                OnServiceStarted.Invoke();

        }

        internal void ServiceStartedEvent()
        {
            if (OnServiceStarted != null)
                OnServiceStarted.Invoke();
        }


        #endregion

        public void Dispose()
        {
            if (_dispose != null)
            {
                _dispose.Dispose();
                _dispose = null;
            }

        }
    }

    public class MicroService : MicroServiceBase
    {
        public MicroService(int port = 9000, string host = "http://localhost", string serviceDisplayName = null, string serviceName = null)
            : base(port: port, host: host, serviceDisplayName: serviceDisplayName, serviceName: serviceName)
        {
        }

        internal override void Start()
        {
            _dispose = WebApp.Start<StartupBase>(string.Concat(_host, ":", _port));
            ServiceStartedEvent();
        }
    }

    public class MicroService<T> : MicroServiceBase
    {
        public MicroService(int port = 9000, string host = "http://localhost", string serviceDisplayName = null, string serviceName = null)
            : base(port: port, host: host, serviceDisplayName: serviceDisplayName, serviceName: serviceName)
        {
        }


        internal override void Start()
        {
            _dispose = WebApp.Start<T>(string.Concat(_host, ":", _port));
            ServiceStartedEvent();
        }
    }
}
